#include <stdio.h>

#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include "vertexShader.c"
#include "fragmentShader.c"

#include "objectHandler.h"
#include "matrixMath.h"
#include "transformation.h"
#include "wavefrontobj.h"
#include "sceneGraph.h"
#include "shader.h"

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdbool.h>
#include <assert.h>

GLuint program;

#define NUM_TEXTURES 5

#define DAY 0
#define NIGHT 1
#define CLOUDS 2
#define OCEAN 3
#define NORMAL 4

int flipFlag = 1;

bool exitRequested = false;

GLFWwindow* window;

GLfloat aspectRatio = 1.0f;

double timeBetweenUpdates = 0.2f;

double timeSinceUpdate = 0.0f;
int framesSinceUpdate = 0;

GLfloat step = 0.0f;
const GLfloat pi = 3.14159f;

vec3 origin = {0.0f, 0.0f, 0.0f};
vec3 cameraPosition = {0.0f, 2.0f, 1.0f};
vec3 objectPosition = {0.0f, 0.0f, 0.0f};

vec3 cameraTravelPosition = {0.0f, 3.0f, 5.5f};
bool cameraTraveling = true;

GLfloat radius = 1.0f;
mat4 viewingTransformation;

// Define a global scene graph root node
SceneNode* rootNode;

void handleCameraTravel() {
    if (cameraTraveling) {
        vec3 delta;

        vec3Subtract(&delta, &cameraTravelPosition, &cameraPosition);

        if (vec3Length(&delta) < 0.0001) cameraTraveling = false;

        vec3Multiply(&delta, &delta, 0.1);
        vec3Add(&cameraPosition, &cameraPosition, &delta);
    }
}

void travelCamera(vec3 newPosition) {
    cameraTravelPosition = newPosition;
    cameraTraveling = true;
}

/**
 * Input handler for camera movement.
 *  */ 
float chalkboardYPosition = 0.0f;
void handleInputs(double deltaTime) {
    assert(window != NULL);

    if (glfwGetKey(window, GLFW_KEY_1) == GLFW_PRESS) {
        travelCamera((vec3){0.0f, 1.7f, 2.4f});
    }
    if (glfwGetKey(window, GLFW_KEY_2) == GLFW_PRESS) {
        travelCamera((vec3){3.3f, 3.4f, -11.0f});
    }
    if (glfwGetKey(window, GLFW_KEY_3) == GLFW_PRESS) {
        travelCamera((vec3){-3.0f, 2.9f, -7.5f});
    }
    if (glfwGetKey(window, GLFW_KEY_4) == GLFW_PRESS) {
        travelCamera((vec3){-0.6f, 2.1f, -4.5f});
    }
    if (glfwGetKey(window, GLFW_KEY_5) == GLFW_PRESS) {
        travelCamera((vec3){-10.6f, 22.1f, -4.5f});
    }
    if (glfwGetKey(window, GLFW_KEY_6) == GLFW_PRESS) {
        travelCamera((vec3){3.574698f, 0.966039f, 3.852249f});
    }
    if (!cameraTraveling) {
        if (glfwGetKey(window, GLFW_KEY_S) == GLFW_PRESS) {
            vec3 delta;

            vec3Subtract(&delta, &origin, &cameraPosition);
            vec3Normalise(&delta, &delta);
            vec3Multiply(&delta, &delta, -10);
            vec3Multiply(&delta, &delta, deltaTime);

            vec3Add(&cameraPosition, &cameraPosition, &delta);
        }
        if (glfwGetKey(window, GLFW_KEY_W) == GLFW_PRESS) {
            vec3 delta;
            vec3Subtract(&delta, &origin, &cameraPosition);
            if (vec3Length(&delta) > 0.1) {
                vec3Normalise(&delta, &delta);
                vec3Multiply(&delta, &delta, 10);
                vec3Multiply(&delta, &delta, deltaTime);

                vec3Add(&cameraPosition, &cameraPosition, &delta);
            }
            
        }
        if (glfwGetKey(window, GLFW_KEY_D) == GLFW_PRESS) {
            vec4 delta;
            mat4 transform;

            vec3Subtract((vec3*)&delta, &origin, &cameraPosition);
            delta.y = 0;
            vec3Normalise((vec3*)&delta, (vec3*)&delta);
            vec3Multiply((vec3*)&delta, (vec3*)&delta, -10);

            rotationY(&transform, pi / 2);
            multiplyAny((GLfloat*)&delta, (GLfloat*)&transform, (GLfloat*)&delta, 4, 4, 1);

            vec3Multiply((vec3*)&delta, (vec3*)&delta, deltaTime);

            vec3Add(&cameraPosition, &cameraPosition, (vec3*)&delta);
        }
        if (glfwGetKey(window, GLFW_KEY_A) == GLFW_PRESS) {
            vec4 delta;
            mat4 transform;

            vec3Subtract((vec3*)&delta, &origin, &cameraPosition);
            delta.y = 0;
            vec3Normalise((vec3*)&delta, (vec3*)&delta);
            vec3Multiply((vec3*)&delta, (vec3*)&delta, 10);

            rotationY(&transform, pi / 2);
            multiplyAny((GLfloat*)&delta, (GLfloat*)&transform, (GLfloat*)&delta, 4, 4, 1);

            vec3Multiply((vec3*)&delta, (vec3*)&delta, deltaTime);

            vec3Add(&cameraPosition, &cameraPosition, (vec3*)&delta);
        }
        if (glfwGetKey(window, GLFW_KEY_F) == GLFW_PRESS || glfwGetKey(window, GLFW_KEY_LEFT_SHIFT) == GLFW_PRESS) {
            cameraPosition.y -= deltaTime * 10;
        }
        if (glfwGetKey(window, GLFW_KEY_R) == GLFW_PRESS || glfwGetKey(window, GLFW_KEY_SPACE) == GLFW_PRESS) {
            cameraPosition.y += deltaTime * 10;
        }
    }

    SceneNode* chalkboard = findNodeByName("myChalkboard1", rootNode);
    if (!chalkboard) return;
    if (glfwGetKey(window, GLFW_KEY_X) == GLFW_PRESS) {
        chalkboardYPosition += deltaTime * 1.0f;
        chalkboardYPosition = fminf(chalkboardYPosition, 1.0f);
        chalkboard->transformation = (mat4){1.0f, 0.0f, 0.0f, 0.0f,
                                        0.0f, 1.0f, 0.0f, 0.0f,
                                        0.0f, 0.0f, 1.0f, 0.0f,
                                        0.0f, chalkboardYPosition, 0.0f, 1.0f};
        updateSceneNode(chalkboard, &findNodeByName("myChalkboard", rootNode)->worldTransformation);
    }
    if (glfwGetKey(window, GLFW_KEY_C) == GLFW_PRESS) {
        chalkboardYPosition -= deltaTime * 1.0f;
        chalkboardYPosition = fmaxf(chalkboardYPosition, 0.0f);
        chalkboard->transformation = (mat4){1.0f, 0.0f, 0.0f, 0.0f,
                                        0.0f, 1.0f, 0.0f, 0.0f,
                                        0.0f, 0.0f, 1.0f, 0.0f,
                                        0.0f, chalkboardYPosition, 0.0f, 1.0f};
        updateSceneNode(chalkboard, &findNodeByName("myChalkboard", rootNode)->worldTransformation);
    }
    
    if (glfwGetKey(window, GLFW_KEY_L) == GLFW_PRESS) {
        objectPosition.x += deltaTime * 10;
    }
    if (glfwGetKey(window, GLFW_KEY_J) == GLFW_PRESS) {
        objectPosition.x -= deltaTime * 10;
    }
    if (glfwGetKey(window, GLFW_KEY_I) == GLFW_PRESS) {
        objectPosition.z += deltaTime * 10;
    }
    if (glfwGetKey(window, GLFW_KEY_K) == GLFW_PRESS) {
        objectPosition.z -= deltaTime * 10;
    }
    if (glfwGetKey(window, GLFW_KEY_O) == GLFW_PRESS) {
        radius += deltaTime * 10;
    }
    if (glfwGetKey(window, GLFW_KEY_U) == GLFW_PRESS) {
        radius -= deltaTime * 10;
    }
}

// input handler to quit with ESC
void keyboardHandler(GLFWwindow* window, int key, int scancode, int action, int mods) {
    assert(window != NULL);
    if (action == GLFW_PRESS) {
        if (key == GLFW_KEY_ESCAPE) {
            exitRequested = true;
        }
    }
}

void renderNode(SceneNode* node) {
    assert(node != NULL);

    if (!node->model) return;

    mat4 modelView;
    identity(&modelView);
    multiply(&modelView, &node->worldTransformation, &modelView);
    multiply(&modelView, &viewingTransformation, &modelView);

    // calculate matrix for normals
    mat3 normalModelView;
    mat3From4(&normalModelView, &modelView);
    mat3Inverse(&normalModelView, &normalModelView);
    mat3Transpose(&normalModelView, &normalModelView);

    // send transformation matrix to shader
    glUniformMatrix4fv(glGetUniformLocation(program, "modelView"), 1, GL_FALSE, (GLfloat*)&modelView);
    glUniformMatrix3fv(glGetUniformLocation(program, "normalModelView"), 1, GL_FALSE, (GLfloat*)&normalModelView);

    // SET MATERIAL DATA
    glUniform1f(glGetUniformLocation(program, "shininess"), node->model->shininess * 4.0f);


    // BIND TEXTURES
    GLuint textureLocation;
    textureLocation = glGetUniformLocation(program, "textureSampler");
    glUniform1i(textureLocation, 0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, node->model->texture);

    if (node->model->secondaryTexture != -1) {

        glUniform1i(glGetUniformLocation(program, "useSecondaryTexture"), 1);

        glUniform1i(glGetUniformLocation(program, "secondaryTexture"), 1);
        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_2D, node->model->secondaryTexture);

    } else {
        glUniform1i(glGetUniformLocation(program, "useSecondaryTexture"), 0);
    }

    if (node->model->normalMap != -1) {

        glUniform1i(glGetUniformLocation(program, "useNormalMap"), 1);

        glUniform1i(glGetUniformLocation(program, "normalMap"), 2);
        glActiveTexture(GL_TEXTURE2);
        glBindTexture(GL_TEXTURE_2D, node->model->normalMap);

    } else {
        glUniform1i(glGetUniformLocation(program, "useNormalMap"), 0);
    }

    draw_object(node->model->objectData);
}

void init(void) {
    //IDEE: shader in array speichern und dann so kompilieren etc
    GLint status;

    Shader shaders[2] = {
        {GL_VERTEX_SHADER, 0},
        {GL_FRAGMENT_SHADER, 0}
    };

    const GLchar* shaderSources[2] = {
        vertexShader_glsl,
        fragmentShader_glsl
    };

    int numShaders = sizeof(shaders) / sizeof(Shader);
    compileShaders(shaders, shaderSources, numShaders);

    program = glCreateProgram();

    // das müsste evtl noch angepasst werden
    for (int i = 0; i < numShaders; i++) {
        glAttachShader(program, shaders[i].id);
    }
    glLinkProgram(program);


    glGetProgramiv(program, GL_LINK_STATUS, &status);

    if (!status) {
        printf("Error linking program: ");
        GLchar infoLog[1024];
        glGetProgramInfoLog(program, 1024, NULL, infoLog);
        printf("%s",infoLog);
        exit(1);
    }
    glValidateProgram(program);


    glGetProgramiv(program, GL_VALIDATE_STATUS, &status);

    if (!status) {
        printf("Error validating program: ");
        GLchar infoLog[1024];
        glGetProgramInfoLog(program, 1024, NULL, infoLog);
        printf("%s",infoLog);
        exit(1);
    }

    stbi_set_flip_vertically_on_load(flipFlag);
    // --------------- READ SCENE GRAPH
    setNodeRenderFunction(&renderNode);

    // read scene graph
    rootNode = loadSceneGraphFromFile("../scene-graph.scg");

    // ENABLE BACKFACE CULLING
    glFrontFace(GL_CCW);
    glEnable(GL_CULL_FACE);

    // ENABLE DEPTH BUFFER
    glEnable(GL_DEPTH_TEST);

    glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
}

void updateStats() {
    printf("\rFPS: %.1f", framesSinceUpdate / timeSinceUpdate);
    printf(" - Camera Position: [%f, %f, %f] - Traveling: %s   ", cameraPosition.x, cameraPosition.y, cameraPosition.z, cameraTraveling ? "true" : "false");
    fflush(stdout);
}

/**
 * Main draw function.
 */
void draw(void) {

    handleCameraTravel();

    // FPS Counter
    framesSinceUpdate++;
    double deltaTime = glfwGetTime();
    timeSinceUpdate += deltaTime;
    glfwSetTime(0.0f);


    if (timeSinceUpdate >= timeBetweenUpdates) {
        updateStats();
        timeSinceUpdate = 0.0f;
        framesSinceUpdate = 0;
    }


    // camera movement
    handleInputs(deltaTime);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glUseProgram(program);

    // step for rotations
    // counts up to 1.0 and then resets back to 0.0 forever
    step += deltaTime / 15;
    if (step > 1.0f) step -= 1.0f;
    if (step < 0.0f) step += 1.0f;

    //SceneNode* box3 = findNodeByName("box3", rootNode);
    //rotateY(&box3->transformation, &box3->transformation, 0.001);
    //updateSceneNode(rootNode, NULL);

    // step multiplied by pi * 2 for use in rotation and trig functions
    GLfloat stepi = step * pi * 2;

    // ------------- VIEWING TRANSFORMATION -------------------
    vec3 up = {0.0f, 1.0f, 0.0f};

    lookAt(&viewingTransformation, &cameraPosition, &origin, &up);

    // -------------- PROJECTION TRANSFORMATION ----------------
    mat4 projectionTransformation;
    GLfloat near = 0.1f;
    GLfloat far = 40.0f;
    perspectiveProjection(&projectionTransformation, near, far);

    // -------------- NORMALISATION TRANSFORMATION -------------
    mat4 normalisationTransformation;
    GLfloat fovy = pi / 2;
    normalisedDeviceCoordinatesFov(&normalisationTransformation, fovy, aspectRatio, near, far);

    mat4 projection;
    identity(&projection);
    multiply(&projection, &projectionTransformation, &projection);
    multiply(&projection, &normalisationTransformation, &projection);

    glUniformMatrix4fv(glGetUniformLocation(program, "projection"), 1, GL_FALSE, (GLfloat*)&projection);

    // SET LIGHT DATA
    glUniform4f(glGetUniformLocation(program, "lightColor"), 0.2f, 0.2f, 0.2f, 0.5f);
    glUniform4f(glGetUniformLocation(program, "ambientLight"), 0.05f, 0.05f, 0.05f, 1.0f);

    vec4 lightPosition = {cos(stepi) * 5.0f, 5.0f, sin(stepi) * 5.0f, 1.0f};
    multiplyAny((GLfloat*)&lightPosition, (GLfloat*)&viewingTransformation, (GLfloat*)&lightPosition, 4, 4, 1);

    glUniform3f(glGetUniformLocation(program, "lightPositions[0]"), lightPosition.x, lightPosition.y, lightPosition.z);
    glUniform3f(glGetUniformLocation(program, "lightPositions[1]"), -lightPosition.x, lightPosition.y, -lightPosition.z);

    glUniform1f(glGetUniformLocation(program, "lightBrightness[0]"), 0.5f);
    glUniform1f(glGetUniformLocation(program, "lightBrightness[1]"), 0.5f);


    renderSceneNode(rootNode);

}

/**
 *  Changes viewport size and adjust aspect ratio when changing window size
 */
void framebuffer_size_callback(GLFWwindow *window, int width, int height) {
    glViewport(0, 0, width, height);
    aspectRatio = (float)width / height;
}

/**
 * Main function.
 */
int main(int argc, char **argv) {
    // initialise window
    glfwInit();
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    window = glfwCreateWindow(700, 700, "Computergrafik 1", NULL, NULL);

    if (!window) {
        printf("Failed to create window\n");
        glfwTerminate();
        return -1;
    }

    glfwSetFramebufferSizeCallback(window, framebuffer_size_callback);
    glfwMakeContextCurrent(window);

    // disable framerate cap
    glfwSwapInterval(0);

    // register keyboard event handler
    glfwSetKeyCallback(window, keyboardHandler);

    // initialise glew
    glewInit();

    printf("OpenGL version supported by this platform (%s):\n", glGetString(GL_VERSION));

    init();
    // exit when window should close or exit is requested (ESC)
    while (!glfwWindowShouldClose(window) && !exitRequested) {
        draw();

        glfwSwapBuffers(window);
        glfwPollEvents();
    }

    glfwTerminate();

    return 0;
}
